
# MMAchords.py

"""
This module is an integeral part of the program 
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bvdp@uniserve.com>

"""


from MMAcommon import *
import copy

"""
Table of chords. All are based on a C scale.
Generating chords is easy in MIDI since we just need to
add/subtract constants, based on yet another table.

CAUTION, if you add to this table make sure there are at least
3 notes in each chord! Don't make any chord longer than 8 notes
(The pattern define sets volumes for only 8).

There is a corresponding scale set for each chord. These are
used by bass and scale patterns.

Each chord needs an English doc string. This is extracted by the
-Dn option to print a table of chordnames for inclusion in the
reference manual.

"""

C  = 0
Cs = Db = 1
D       = 2
Ds = Eb = 3
E  = Fb = 4
Es = F  = 5
Fs = Gb = 6
G       = 7
Gs = Ab = 8
A  = Bbb= 9
As = Bb = 10
B  = Cb = 11

_chords = {
	'M':	((C,	E,	  G ),
			 (C, D, E, F, G, A, B),
			 "Major triad. This is the default and is used in  "
			 "the absense of any other chord type specification."),

	'm':	((C,	Eb,    G ),
			 (C, D, Eb, F, G, Ab, Bb),
			 "Minor triad."),

	'mb5':	((C,	Eb,	   Gb ),
			 (C, D, Eb, F, Gb, Ab, Bb),
			 "Minor triad with flat 5th."),

	'm#5':	((C,	Eb,	   Gs ),
			 (C, D, Eb, F, Gs, Ab, Bb),
			 "Major triad with augmented 5th."),

	'm6':	((C,	Eb,	   G, A ),
			 (C, D, Eb, F, G, A, Bb),
			 "Minor 6th."),

	'm7':	((C,	Eb,	   G,	  Bb ),
			 (C, D, Eb, F, G, Ab, Bb),
			 "Minor 7th."),

	'mM7':	((C,	Eb,	  G,	  B ),
			 (C, D, Eb, F, G, Ab, B),
			 "Minor Triad plus Major 7th. You will also see this printed "
			 "as ``m(maj7)'', ``m+7'', ``min(maj7)'' and ``min$\sharp$7'' "
			 "(which \mma\ accepts); as well as the \mma\ \emph{invalid} "
			 "forms: ``-($\Delta$7)'', and ``min$\\natural$7''."),
	
	'm7b5': ((C,	Eb,	   Gb,	   Bb ),
			 (C, D, Eb, F, Gb, Ab, Bb),
			 "Minor 7th, flat 5 (aka 1/2 diminished). "),

	'm7b9': ((C,     Eb,    G,     Bb, Db+12 ),
			 (C, Db, Eb, F, G, Ab, Bb),
			 "Minor 7th with added flat 9th."),

	'7':	((C,    E,	  G,	Bb ),
			 (C, D, E, F, G, A, Bb),
			 "Dominant 7th."),

	'7#5':	((C,    E,    Gs,    Bb ),
			 (C, D, E, F, Gs, A, Bb),
			 "7th, sharp 5."),

	'7b5':	((C,    E,	  Gb,    Bb ),
			 (C, D, E, F, Gb, A, Bb),
			 "7th, flat 5."),

	'dim7': ((C,	Eb,	   Gb,	   Bbb ),
			 (C, D, Eb, F, Gb, Ab, Bbb ),	# missing 8th note
			 "Diminished seventh."),

	'aug':	((C,    E,	  Gs ),
			 (C, D, E, F, Gs, A, B ),
			 "Augmented triad."),

	'6':	((C,    E,	  G, A ),
			 (C, D, E, F, G, A, B),
			 "Major tiad with added 6th."),
	
	'M7':	((C,    E,    G,    B),
			 (C, D, E, F, G, A, B),
			 "Major 7th."),

	'M7#5':	((C,    E,    Gs,    B),
			 (C, D, E, F, Gs, A, B),
			 "Major 7th with sharp 5th."),
	
	'M7b5': ((C,    E,	  Gb,    B ),
			 (C, D, E, F, Gb, A, B ),
			 "Major 7th with a flat 5th."),

	'9':	((C,    E,    G,    Bb, D+12 ),
			 (C, D, E, F, G, A, Bb),
			 "Dominant 7th plus 9th."),

	'sus9': ((C,    E,    G,    D+12),
			 (C, D, E, F, G, A, D+12),
			 "Dominant 7th plus 9th, omit 7th."),

	'9b5':	((C,    E,    Gb,    Bb, D+12 ),
			 (C, D, E, F, Gb, A, Bb),
			 "Dominant 7th plus 9th with flat 5th."),

	'm9':	((C,	Eb,	   G,	  Bb, D+12 ),
			 (C, D, Eb, F, G, Ab, Bb),
			 "Minor triad plus 7th and 9th."),

	'm9b5': ((C,	Eb,    Gb, Bb, D+12 ),
			 (C, D, Eb, F, Gb, Ab, Bb),
			 "Minor triad, flat 5, plus 7th and 9th."),

	'm(sus9)':((C,	  Eb,    G,     D+12 ),
			   (C, D, Eb, F, G, Ab, D+12),
			   "Minor triad plus 9th (no 7th)."),

	'M9':	((C,    E,    G,    B, D+12 ),
			 (C, D, E, F, G, A, B),
			 "Major 7th plus 9th."),

	'7b9':	((C,     E,    G,    Bb, Db+12 ),
			 (C, Db, E, F, G, A, Bb),
			 "Dominant 7th with flat 9th."),

	'7#9':	((C,     E,    G,    Bb, Ds+12 ),
			 (C, Ds, E, F, G, A, Bb),
			 "Dominant 7th with sharp 9th."),

	'7b5b9':((C,     E,    Gb,    Bb, Db+12 ),
			 (C, Db, E, F, Gb, A, Bb),
			 "Dominant 7th with flat 5th and flat 9th."),

	'7b5#9':((C,     E,    Gb,    Bb, Ds+12 ),
			 (C, Ds, E, F, Gb, A, Bb),
			 "Dominant 7th with flat 5th and sharp 9th."),

	'7#5#9':((C,     E,    Gs,    Bb, Ds+12 ),
			 (C, Ds, E, F, Gs, A, Bb),
			 "Dominant 7th with sharp 5th and sharp 9th."),

	'7#5b9':((C,     E,    Gs,    Bb, Db+12 ),
			 (C, Db, E, F, Gs, A, Bb),
			 "Dominant 7th with sharp 5th and flat 9th."),

	'aug7': ((C,    E,    Gs,    Bb ),
			 (C, D, E, F, Gs, A, Bb),
			 "An augmented chord (raised 5th) with a dominant 7th."),

	'aug7b9':((C,     E,    Gs,    Bb, Db+12 ),
			  (C, Db, E, F, Gs, A, Bb),
			  "Augmented 7th with flat 5th and sharp 9th."),

	'11':	((C,    E,    G,    Bb, D+12, F+12 ),
			 (C, D, E, F, G, A, Bb),
			 "9th chord plus 11th."),

	'm11':	((C,    Eb,    G,     Bb, D+12, F+12 ),
			 (C, D, Eb, F, G, Ab, Bb),
			 "9th with minor 3rd,  plus 11th."),

	'11b9': ((C,     E,    G,    Bb, Db+12, F+12 ),
			 (C, Db, E, F, G, A, Bb),
			 "9th chord plus flat 11th."),

	'9#5':	((C,    E,    Gs,    Bb, D+12 ),
			 (C, D, E, F, Gs, A, Bb),
			 "Dominant 7th plus 9th with sharp 5th."),

	'9#11': ((C,    E,     G,    Bb, D+12, Fs+12 ),
			 (C, D, E, Fs, G, A, Bb),
			 "Dominant 7th plus 9th and sharp 11th."),

	'7#9#11':((C,     E,     G,    Bb, Ds+12, Fs+12 ),
			  (C, Ds, E, Fs, G, A, Bb),
			  "Dominant 7th plus sharp 9th and sharp 11th."),


	'M7#11':((C,    E,     G,    B, D+12, Fs+12 ),
			 (C, D, E, Fs, G, A, B),
			 "Major 7th plus 9th and sharp 11th."),
	
	# Sus chords. Not sure what to do with the associated scales. For
	# now just duplicating the 2nd or 3rd in the scale seems to make sense.

	'sus4': ((C,    F,    G ),
			 (C, D, F, F, G, A, B),
			 "Suspended 4th, major triad with 3rd raised half tone."),
	
	'7sus': ((C,    F,    G,    Bb ),
			 (C, D, F, F, G, A, Bb),
			 "7th with suspended 4th, dominant 7th with 3rd "
			 "raised half tone."),

	'sus2': ((C,    D,    G ),
			 (C, D, D, F, G, A, B),
			 "Suspended 2nd, major triad with major 2nd above "
			 "root substituted for 3rd."),

	'7sus2':((C,    D,    G,    Bb ),
			 (C, D, D, F, G, A, Bb),
			 "A sus2 with dominant 7th added."),
	
	# these two chords should probably NOT have the 5th included,
	# but since a number of voicings depend on the 5th being
	# the third note of the chord, they're here.

	'13':	((C,    E,    G,    Bb, A+12),
			 (C, D, E, F, G, A, Bb),
			 "Dominant 7th (including 5th) plus 13th."),

	'M13':	((C,    E,    G,    B, A+12),
			 (C, D, E, F, G, A, B),
			 "Major 7th (including 5th) plus 13th."),

	# Because some patterns assume that the 3rd note in a chord is a 5th,
	# or a varient, we duplicate the root into the position of the 3rd ... and
	# to make the sound even we duplicate the 5th into the 4th position as well.
	
	'5':	((C, C,	G, G ),
			 (C, D, E, F, G, A, B),
			 "Altered Fifth or Power Chord; root and 5th only."),
}


""" Extend our table with common synomyns. These are real copies,
not pointers. This is done so that a user redefine only affects
the original.
"""

for a,b,d in (
	('aug9',     '9#5'  , ''),
	('9+5',      '9#5'  , ''),
	('m+5',		 'm#5'	, ''),
	('M6',		 '6'	, ''),
	('m7-5',	 'm7b5' , ''),
	('+',		 'aug'	, ''),
	('+7',       'aug7' , ''),
	('#5',       'aug'  , ''),
	('7-9',		 '7b9'	, ''),
	('7+9',		 '7#9'	, ''),
	('maj7',	 'M7'	, ''),
	('M7-5',	 'M7b5'	, ''),
	('M7+5',	 'M7#5'	, ''),
	('7sus4',	 '7sus' , ''),
	('7#11',	 '9#11' , ''),
	('7+',		 'aug7' , ''),
	('7+5',		 '7#5'	, ''),
	('7-5',		 '7b5'	, ''),
	('sus',		 'sus4' , ''),
	('m(maj7)',	 'mM7'	, ''),
	('m+7',		 'mM7'	, ''),
	('min(maj7)','mM7'	, ''),
	('min#7',	 'mM7'	, ''),
	('dim',		 'dim7' , 'A dim7, not a triad!'),
	('9sus',	 'sus9' , ''),
	('9-5',      '9b5'  , ''),
	('dim3',	 'mb5'	, 'Diminished triad (non-standard notation).') ):

	n=_chords[b][0]
	s=_chords[b][1]
	if not d:
		d=_chords[b][2]

	_chords[a] = (n, s, d)



def defChord(ln):
	""" Add a new chord type to the _chords{} dict. """

	emsg="DefChord needs NAME (NOTES) (SCALE)"

	# At this point ln is a list. The first item should be
	# the new chord type name.

	if not len(ln):
		error(emsg)
	name = ln.pop(0)
	if name in _chords.keys():
		warning("Redefining chordtype '%s'." % name)

	if name.count('/'):
		error("A slash in not permitted in chord type name")

	ln=pextract(''.join(ln), '(', ')')

	if ln[0] or len(ln[1])!=2:
		error(emsg)

	notes=ln[1][0].split(',')
	if len(notes) < 2 or len(notes)>8:
		error("There must be 2..8 notes in a chord, not '%s'." % len(note))
	notes.sort()
	for i,v in enumerate(notes):
		v=stoi(v, "Note offsets in chord must be integers, not '%s'." % v)
		if v<0 or v>24:
			error("Note offsets in chord must be 0..24, not '%s'." % v)
		notes[i]=v

	scale=ln[1][1].split(',')
	if len(scale) != 7:
		error("There must be 7 offsets in chord scale, not '%s'" % len(scale))
	scale.sort()
	for i,v in enumerate(scale):
		v=stoi(v, "Scale offsets in chord must be integers, not '%s'." % v)
		if v<0 or v>24:
			error("Scale offsets in chord must be 0..24, not '%s'." % v)
		scale[i]=v


	_chords[name] = ( notes, scale, "User Defined") 

	if gbl.debug:
		print "ChordType '%s', %s" % (name, _chords[name])


def printChord(ln):
	""" Display the note/scale/def for chord(s). """

	for c in ln:
		if not _chords.has_key(c):
			error("Chord '%s' is unknown" % c)
		print c, ':', _chords[c][0], _chords[c][1], _chords[c][2]


"""
Table of chord adjustment factors. Since the initial chord is based
on a C scale, we need to shift the chord for different degrees. Note,
that with C as a midpoint we shift left for G/A/B and right for D/E/F.

Should the shifts take in account the current key signature?
"""

_chordAdjust = {
	'Gb':-6,
	'G' :-5,
	'G#':-4, 'Ab':-4,
	'A' :-3,
	'A#':-2, 'Bb':-2,
	'B' :-1, 'Cb':-1,
	'B#': 0, 'C' : 0,
	'C#': 1, 'Db': 1,
	'D' : 2,
	'D#': 3, 'Eb': 3,
	'E' : 4, 'Fb': 4,
	'E#': 5, 'F' : 5,
	'F#': 6 }

def chordAdjust(ln):
	""" Adjust the chord point up/down one octave. """

	if not ln:
		error("ChordAdjust: Needs at least one argument.")

	for l in ln:
		try:
			pitch, octave = l.split('=')
		except:
			error("Each arg must contain an '=', not '%s'." % l)
			
		if pitch not in _chordAdjust:
			error("ChordAdjust: '%s' is not a valid pitch." % pitch)

		octave = stoi(octave, "ChordAdjust: expecting integer, not '%s'." % octave)

		p=_chordAdjust[pitch]
		if octave == 0:
			if p < -6:
				_chordAdjust[pitch] += 12
			elif p > 6:
				_chordAdjust[pitch]-=12
				
		elif octave == -1 and p <= 6 and p >= -6:
			_chordAdjust[pitch] -= 12

		elif octave == 1 and p <= 6 and p >= -6:
			_chordAdjust[pitch] += 12

		else:
			error("ChordAdjust: '%s' is not a valid octave. Use 1, 0 or -1." % octave)


	
###############################
# Chord creation/manipulation #
###############################

class ChordNotes:
	""" The Chord class creates and manipulates chords for MMA. The
	class is initialized with a call with the chord name. Eg:

	ch = ChordNotes("Am")

	The following methods and variables are defined:

	noteList  - the notes in the chord as a list. The "Am"
		would be [9, 12, 16].

	noteListLen	 - length of noteList.

	tonic	   - the tonic of the chord ("Am" would be "A").

	chordType  - the type of chord ("Am" would be "m").

	rootNote   - the root note of the chord ("Am" would be a 9).

	bnoteList  - the original chord notes, bypassing any
	invert(), etc. mangling.

	scaleList  - a 7 note list representing a scale similar to
	the chord.

	reset() - resets noteList to the original chord notes.
	This is useful to restore the original after
	chord note mangling by invert(), etc. without having to
	create a new chord object.


	invert(n) - Inverts a chord by 'n'. This is done inplace and
	returns None. 'n' can have any integer value, but -1 and 1
	are most common. The order of the notes is not changed. Eg:

	ch=Chord('Am') 
	ch.noteList == [9, 12, 16]
	ch.invert(1)
	ch.noteList	 = [21, 12, 16]				

	compress() - Compresses the range of a chord to a single octave. This is
	done inplace and return None. Eg:

	ch=Chord("A13")
	ch.noteList == [1, 5, 8, 11, 21]
	ch.compress()
	ch.noteList == [1, 5, 8, 11, 10 ]


	limit(n) -	Limits the range of the chord 'n' notes. Done inplace
	and returns None. Eg:

	ch=Chord("CM711")
	ch.noteList == [0, 4, 7, 11, 15, 18]   
	ch.limit(4)
	ch.noteList ==	[0, 4, 7, 11]
	
	
	"""


	#################
	### Functions ###
	#################

	def __init__(self, name):
		""" Create a chord object. Pass the chord name as the only arg. 

		NOTE: Chord names ARE case-sensitive!

		The chord NAME at this point is something like 'Cm' or 'A#7'. 
		Split off the tonic and the type. 
		If the 2nd char is '#' or 'b' we have a 2 char tonic,
		otherwise, it's the first char only.
		
		A chord can start with a single '+' or '-'. This moves
		the entire chord and scale up/down an octave.

		Note pythonic trick: By using ranges like [1:2] we
		avoid runtime errors on too-short strings. If a 1 char
		string,	 name[1] is an error; name[1:2] just returns None.

		Further note: I have tried to enable caching of the generated
		chords, but found no speed difference. So, to make life simpler
		I've decided to generate a new object each time.
		
		"""

		slash = None
		octave = 0

		if name == 'z':
			self.tonic = self.chordType = None
			self.noteListLen = 0
			self.notesList = self.bnoteList = []
			return

		if name.startswith('-'):
			name = name[1:]
			octave = -12

		if name.startswith('+'):
			name = name[1:]
			octave = 12

		name = name.replace('&', 'b')

		# Strip off the slash part of the chord. Use later
		# to do proper inversion.

		if name.find('/') > 0:
			name, slash = name.split('/')

		if name[1:2] in ( '#b' ):
			tonic = name[0:2]			
			ctype  = name[2:]
		else:
			tonic = name[0:1]
			ctype  = name[1:]

		if not ctype:		# If no type, make it a Major
			ctype='M'

		try:
			notes = _chords[ctype][0]
			adj =	_chordAdjust[tonic] + octave
		except:
			error( "Illegal/Unknown chord in '%s'. Tonic='%s', Type='%s'" \
				   % (name, tonic, ctype) )

		## out...this cause some undesired octave jumps.
		##if notes[0] + adj < 0:
		##	adj += 12

		self.noteList	 = [ x + adj for x in notes ]
		self.bnoteList	 = tuple(self.noteList)
		self.scaleList	 = tuple([ x + adj for x in _chords[ctype][1] ])
		self.chordType	 = ctype
		self.tonic		 = tonic
		self.rootNote	 = self.noteList[0]

		self.noteListLen = len(self.noteList)

		# Do inversions if there is a valid slash notation.

		if slash:

			if not _chordAdjust.has_key(slash):
				error("The note '%s' in the slash chord is unknown." % slash)

			r=_chordAdjust[slash]	# r = -6 to 6

			# If the slash note is in the chord we invert
			# the chord so the slash note is in root position.

			c_roted = 0
			s=self.noteList
			for octave in [0, 12, 24]:
				if r+octave in s:
					rot=s.index(r+octave)
					for i in range(rot):
						s.append(s.pop(0)+12)
					if s[0] >= 12:
						for i,v in enumerate(s):
							s[i] = v-12
							self.noteList = s
					self.bnoteList = tuple(s)
					self.rootNote = self.noteList[0]
					c_roted = 1
					break

			s_roted = 0
			s=list(self.scaleList)
			for octave in [0, 12, 24]:
				if r+octave in s:
					rot=s.index(r+octave)
					for i in range(rot):
						s.append(s.pop(0)+12)
						if s[0] > 12:
							for i,v in enumerate(s):
								s[i] = v-12
						self.scaleList=tuple(s)
					s_roted = 1
					break

			if not c_roted and not s_roted:
				warning("The slash chord note '%s' not in "
						"chord or scale." % slash)

			elif not c_roted:	
				warning("The slash chord note '%s' not in "
						"chord '%s'" % (slash, name))

			elif not s_roted:	# Probably will never happen :)
				warning("The slash chord note '%s' not in "
						"scale for the chord '%s'" % (slash, name))


	def reset(self):
		""" Restores notes array to original, undoes mangling. """

		self.noteList	 = list(self.bnoteList[:])
		self.noteListLen = len(self.noteList)


	def invert(self, n):
		""" Apply an inversion to a chord.

		This does not reorder any notes, which means that the root note of
		the chord reminds in postion 0. We just find that highest/lowest
		notes in the chord and adjust their octave.
		
		NOTE: Done on the existing list of notes. Returns None.
		"""

		if n:
			c=self.noteList[:]

			while n>0:		# Rotate up by adding 12 to lowest note
				n -= 1
				c[c.index(min(c))]+=12

			while n<0:		# Rotate down, subtract 12 from highest note
				n += 1
				c[c.index(max(c))]-=12

			self.noteList = c

		return None 



	def compress(self):
		""" Compress a chord to one ocatve.
		
		
		Get max permitted value. This is the lowest note
		plus 12. Note: use the unmodifed value bnoteList!
		"""

		mx = self.bnoteList[0] + 12		
		c=[]

		for i, n in enumerate(self.noteList):
			if n > mx:
				n -= 12
			c.append(n)

		self.noteList = c

		return None



	def limit(self, n):
		""" Limit the number of notes in a chord. """

		if n < self.noteListLen:
			self.noteList =	 self.noteList[:n]
			self.noteListLen = len(self.noteList)

		return None


	def center1(self, lastChord):
		""" Descriptive comment needed here!!!! """

		def minDistToLast(x, lastChord):
			dist=99
			for j in range(len(lastChord)):
				if abs(x-lastChord[j])<abs(dist):
					dist=x-lastChord[j]
			return dist

		def sign(x):
			if (x>0):
				return 1
			elif (x<0):
				return -1
			else:
				return 0			

		# Only change what needs to be changed compared to the last chord
		# (leave notes where they are if they are in the new chord as well).

		if lastChord:
			ch=self.noteList

			for i in range(len(ch)):

				# minimize distance to last chord

				oldDist = minDistToLast(ch[i], lastChord)				
				while abs(minDistToLast(ch[i] - sign(oldDist)*12,
								lastChord)) < abs(oldDist):
					ch[i] -= 12* sign(oldDist) 
					oldDist = minDistToLast(ch[i], lastChord)

		return None

	def center2(self, centerNote, noteRange):
		""" Need COMMENT """

		ch=self.noteList
		for i,v in enumerate(ch):

			dist = v - centerNote 
			if dist < -noteRange:
				ch[i] = v + 12 * ( abs(dist) / 12+1 )
			if dist > noteRange:
				ch[i] = v - 12 * ( abs(dist) / 12+1 )

	
		return None


######## End of Chord class #####



def docs():
	""" Print out a list of chord names and docs in LaTex. """

	import copy

	# Just in case someone else wants to use _chords, work on a copy

	chords=copy.copy(_chords)

	a=chords.keys()
	a.sort()
	for n in a:
		nm=n.replace("#", '$\\sharp$')
		nm=nm.replace('b', '$\\flat$')
		print "\\insline{%s}{%s}" % (nm, chords[n][2])



